#!/bin/bash
COPYRIGHT='(c) wiimm (at) wiimm.de -- 2020-02-02'

export LC_ALL=C

export COL_LOG="\033[44;37;1m"
export COL_SUM="\033[33;1m"
export COL_ACTIVITY="\033[36;1m"
export COL_ERROR="\033[41;37;1m"
export COL0="\033[0m"

printf "\n\n\n${COL_SUM}Setup... ${COL0}\n\n"
script="${0##*/}"

# pre definition
ERR_WARN=28
ERR_ERROR=112

#
#------------------------------------------------------------------------------
# system and bin path

#--- BASEDIR

if [[ ${BASH_SOURCE:0:1} == / ]]
then
    BASEDIR="${BASH_SOURCE%/*}"
    [[ $BASEDIR = "" ]] && BASEDIR=/
else
    BASEDIR="$PWD/$BASH_SOURCE"
    BASEDIR="${BASEDIR%/*}"
fi
#echo "0=$0"
#echo "BASH_ARGV=${BASH_ARGV[*]}"
export BASEDIR


#--- predefine BINDIR & PATH for Cygwin

export ORIGPATH="$PATH"
export BINDIR="$BASEDIR/cygwin"
[[ -d $BINDIR ]] && export PATH="$BINDIR:$ORIGPATH"


#--- find system

export SYSTEM="$( uname -s | tr '[A-Z]' '[a-z]' )"
export MACHINE="$( uname -m | tr '[A-Z]' '[a-z]' )"
export HOST

case "$SYSTEM-$MACHINE" in
    darwin-*)		HOST=mac ;;
    linux-x86_64)	HOST=linux64 ;;
    linux-*)		HOST=linux32 ;;
    cygwin*)		HOST=cygwin ;;
    *)			HOST=- ;;
esac


#--- setup BINDIR and PATH

BINDIR="$BASEDIR/$HOST"
((VERBOSE>0)) && echo "BINDIR      = $BINDIR"
if [[ -d $BINDIR ]]
then
    #[[ $HOST = cygwin ]] || chmod u+x "$BINDIR"/* 2>/dev/null
    chmod u+x "$BINDIR"/* 2>/dev/null || true
    export PATH="$BINDIR:$ORIGPATH"
fi


#--- misc

export workdir=./patch-dir

export LOGDIR="$PWD/_log"
mkdir -p "$LOGDIR"
chmod 777 "$LOGDIR" 2>/dev/null

#
#------------------------------------------------------------------------------
# print help

function help_exit()
{
    local sed='/^%/d; s/^\|//'
    [[ $1 = x ]] && sed='s/^%\t?//; s/^\|//'
    sed -r "$sed" <<- __EOT__

	Usage: $script [option]...

	Options:

	  -h  --help          : Print this help and exit.
	  -t  --test          : Enter test mode => do nothing
	  -v  --verbose       : Be verbose, or more verbose
	  -q  --quiet         : Be quiet, disable --verbose
	  -a  --autorun       : Read definition file and run without interaction
	  -f  --fast          : Enable fast mode (weak compresion)
	  -o  --optfile       : Load option file 'mkf-options.sh' after extracting
	      --local         : Use wit & wszst from /usr/local/bin if exists and newer
	  -3  --patch3        : Extract patch3 if available

	  -S  --src file      : Short cut for --source
	      --source path   : DIR:  first try this directory as source
	 			FILE: use this file as source
	  -D  --dest path     : if DIR:  create new image here and not in ./new-image/
	|			if FILE: use this as path and name for new image

	      --riiv-mode md  : supported riivolution mode
	|			0=auto, 1:MKW full patch, 2:MKW, 3:other games

	  -N  --none          : patch only and don't create an image.
	      --wbfs          : set output to *.wbfs files (default)
	      --wdf           : set output to *.wdf  files
	      --wia           : set output to *.wia  files
	      --gcz           : set output to *.gcz  files
	      --iso           : set output to *.iso  files
	      --ciso          : set output to *.ciso files
	      --riiv          : create Riivolution setup instead of image
	      --riiv-         : like --riiv, but without patching any source file
	      --add           : call: wwt add -aovv
	|			(wwt is not part of this distribution)

	      --split         : split output at 4 GB
	  -s  --savegame      : change ID to 'K.....' to enable separate savegame
%
%	Developer Options:
%
%	      --xhelp	      : Print this extended help.
%	      --relax	      : Ignored.
%	  -d  --distrib	      : Create a distribution archive.
%				If set twice: Stop execution after creating.
%	  -0  --nopack	      : Neither create *.txz nor *.7z.
%	  -r  --rsync	      : Rsync distrib to server 'download.wiimm.de'.
%				If set twice: Build *.txz
%				If set 3 times: Build *.txz and get it.
%	  -x  --xz	      : Create *.txz (tar.xz) archive. Default if -7 not set.
%	  -7  --7z	      : Create a *.7z archive. On -7x, create both archives.
%				If set twice: Same as -7xx.
%	  -@  --dbdistrib     : Create a DB based (ct.wiimm.de) distribution.
%				This option includes --distrib.
%
%	       --old	      : Set OPT_NEW := -1 for testing old variants.
%	       --new	      : Set OPT_NEW := +1 for testing new variants.

	$COPYRIGHT

	__EOT__
    exit 1
}

#
#------------------------------------------------------------------------------
# check existence of tools

needed_tools="
	awk bash cat chmod cp cut date diff find grep ln
	mkdir mv rm sed sort tar touch tr uname uniq unzip wc which
	wit wszst
"

err=

for tool in $needed_tools
do
    if ! which $tool >/dev/null 2>&1
    then
	err+=" $tool"
    fi
done

if [[ $err != "" ]]
then
    printf "\n\033[31;1m!!! Missing tools:$err => abort!\033[0m\n\n" >&2
    printf "\033[36mPATH:\n   " >&2
    sed 's/:/\n   /g' <<< "$PATH" >&2
    printf "\033[0m\n" >&2
    exit 1
fi

#printf '\e[0;30;47m HOST: %s \e[m\n' "$HOST"

#
#------------------------------------------------------------------------------
# scan options with: wszst getopt

#-- special handling if first param begins with '@' => insert content of file

if [[ $1 =~ ^@ ]]
then
    f="${1:1}"
    shift
    if [[ -f $f ]]
    then
	old=("$@")
	eval set -- $(cat "$f")
	set -- "$@" "${old[@]}"
    fi
fi


#-- scan options

args=$( "$BINDIR/wszst" getopt -n "$script" \
	-l relax,help,xhelp,test,new,old,verbose,quiet,autorun,fast,optfile,local \
	-l patch3,distrib,dbdistrib,nopack,rsync,xz,7z,src:,source:,dest:,riiv-mode: \
	-l none,wbfs,wdf,wdf1,wdf2,wia,gcz,iso,ciso,riiv,riiv-,add,split,savegame \
	-o htvqafo3d@0rx7S:D:Ns \
	-- "$@" )

if (( $? ))
then
    echo "${0##*/}: try option -h or --help for more help" >&2
    exit 1
fi

eval set -- $args

#
#------------------------------------------------------------------------------
# scan options

OPT_IMG_SET=

function setup_image_mode()
{
    mode="$1";
    [[ $mode = -N ]] && mode=none
    [[ $mode =~ ^-- ]] && mode="${mode:2}"

    case "$mode" in
	none)	OPT_IMG_RIIV=0 OPT_IMG_MODE=none ;;

	wbfs)	OPT_IMG_RIIV=0 OPT_IMG_MODE=wbfs ;;
	wdf)	OPT_IMG_RIIV=0 OPT_IMG_MODE=wdf2 ;;
	wdf1)	OPT_IMG_RIIV=0 OPT_IMG_MODE=wdf1 ;;
	wdf2)	OPT_IMG_RIIV=0 OPT_IMG_MODE=wdf2 ;;
	wia)	OPT_IMG_RIIV=0 OPT_IMG_MODE=wia ;;
	gcz)	OPT_IMG_RIIV=0 OPT_IMG_MODE=gcz ;;
	iso)	OPT_IMG_RIIV=0 OPT_IMG_MODE=iso ;;
	ciso)	OPT_IMG_RIIV=0 OPT_IMG_MODE=ciso ;;
	add)	OPT_IMG_RIIV=0 OPT_IMG_MODE=add ;;

	riiv)	((RIIV_ENABLED)) && OPT_IMG_RIIV=1 OPT_IMG_MODE=Riivolution ;;
	riiv-)	((RIIV_ENABLED)) && OPT_IMG_RIIV=2 OPT_IMG_MODE=Riivolution- ;;

	*)	return
    esac

    OPT_IMG_SET="$mode"
}

export TESTMODE=0
export VERBOSE=0
export OPT_AUTORUN=0
export OPT_DISTRIB=0
export OPT_DB_DISTRIB=0
export OPT_PACK=1
export OPT_RSYNC=0
export OPT_XZ=0
export OPT_7Z=0

export SRC_DIR=
export SRC_FILE=
export DEST_FILE=

export OPT_SPLIT=0
export OPT_SAVEGAME=0
export OPT_FAST=0
export OPT_NEW=0
export OPTION_FILE=
export LOCAL_TOOL=0
export OPT_PATCH3=0

export RIIV_SRC="$BASEDIR/riivolution"
export RIIV_ENABLED=0
[[ -d $RIIV_SRC ]] && RIIV_ENABLED=1
export OPT_RIIV_MODE=0
export OPT_IMG_RIIV=0
export OPT_IMG_MODE=wbfs

scan_parm=0
while (($#))
do
    if ((scan_parm))
    then
	help_exit
    else
      #echo "$1"
      case "$1" in
	--)		scan_parm=1 ;;
	   --relax)	;;
	-h|--help)	help_exit ;;
	   --xhelp)	help_exit x ;;
	-t|--test)	let TESTMODE++ ;;
	   --new)	OPT_NEW=$(( OPT_NEW<0 ? +1 : OPT_NEW+1 )) ;;
	   --old)	OPT_NEW=$(( OPT_NEW>0 ? -1 : OPT_NEW-1 )) ;;
	-v|--verbose)	let VERBOSE++ ;;
	-q|--quiet)	VERBOSE=-1 ;;
	-a|--autorun)	let OPT_AUTORUN++;;
	-d|--distrib)	let OPT_DISTRIB++ ;;
	-@|--dbdistrib)	let OPT_DISTRIB++ OPT_DB_DISTRIB++ ;;
	-0|--nopack)	OPT_PACK=0 OPT_XZ=0 OPT_7Z=0 ;;
	-r|--rsync)	let OPT_RSYNC++ ;;
	-x|--xz)	OPT_PACK=1; let OPT_XZ++ ;;
	-7|--7z)	OPT_PACK=1; let OPT_7Z++ ;;
	-f|--fast)	OPT_FAST=1 ;;
	-o|--optfile)	OPTION_FILE=mkf-options.sh ;;
	   --local)	let LOCAL_TOOL++ ;;
	-3|--patch3)	let OPT_PATCH3++ ;;

	--wbfs|--wdf|--wdf1|--wdf2|--wia|--gcz|--iso|--ciso)
			setup_image_mode "$1" ;;
	-N|--none|--add|--riiv|--riiv-)
			setup_image_mode "$1" ;;
	   --riiv-mode)	shift; OPT_RIIV_MODE="$1" ;;

	-S|--src|--source) shift; SRC_DIR="$1" ;;
	-D|--dest)	shift; DEST_FILE="$1" ;;

	   --split)	OPT_SPLIT=1 ;;
	-s|--savegame)	OPT_SAVEGAME=1 ;;

	  *)		echo "${0##*/}: INTERNAL ERROR 'option': $1" >&2; exit 2 ;;
      esac
    fi
    shift
done

#---------------

((OPT_DISTRIB>1)) && OPT_AUTORUN=1	# suppress config queries
((OPT_FAST)) && fast=--fast || fast=

WIT_VERBOSE=-vv
((VERBOSE<0)) && WIT_VERBOSE=-q

WIT_SPLIT=()
((OPT_SPLIT)) && WIT_SPLIT=(--split=4g)

(( OPT_RIIV_MODE > ENABLE_RIIV )) && OPT_RIIV_MODE=$ENABLE_RIIV
(( OPT_IMG_RIIV > 0 && OPT_RIIV_MODE > 2 )) && OPT_IMG_RIIV=$OPT_RIIV_MODE
(( OPT_IMG_RIIV > ENABLE_RIIV )) && OPT_IMG_RIIV=$ENABLE_RIIV

TEST=
((TESTMODE)) && TEST="echo WOULD DO:"

#---------------

function SelectTool() # tool
{
    local awk_ver='{gsub(/\r/,"",$2)} $1=="version" {v="v"$2} $1=="revision" {r=" r"$2} \
		END {print v r}'
    local awk_rev='$1=="revision" {gsub(/\r/,"",$2); print $2; exit}'

    local tool="$1"
    local bin="$BINDIR/$tool"
    local ver="$( "$bin" version -l | awk -F= "$awk_ver" )"
    local rev="$( "$bin" version -l | awk -F= "$awk_rev" )"
    local rev="${rev/[!0-9]*}"
    if ((LOCAL_TOOL))
    then
	local lbin="/usr/local/bin/$tool"
	if [[ -x $lbin ]]
	then
	    local lver="$( "$lbin" version -l | awk -F= "$awk_ver" )"
	    local lrev="$( "$lbin" version -l | awk -F= "$awk_rev" )"
	    local lrev="${lrev/[!0-9]*}"
	    if [[ "$lrev" > "$rev" ]]
	    then
		bin="$lbin"
		ver="$lver"
		rev="$lrev"
	    fi
	fi
    fi

    eval export ${tool^^}=\"$bin\"
    eval export ${tool^^}_VER=\"$ver\"
    eval export ${tool^^}_REV=\"$rev\"
}

SelectTool wit
SelectTool wszst

# get error codes
ERR_WARN=$( "$WSZST" error -B warning )
ERR_ERROR=$( "$WSZST" error -B error )

#---------------

if (( VERBOSE >= 0 ))
then
  sep='--------------------------------------'
  sep="##$sep$sep"$'\n'
  printf '\n%s' "$sep"
  format="## \033[32m%10s\033[0m = \033[36m%.70s\033[0m\n"

  if (( VERBOSE > 0 ))
  then
    ( "$WIT" --version; "$WSZST" --nocolor --version ) | sed 's/^/## /'
    printf "$sep"
    [[ $PATCHER_NAME ]] && printf "$format" PATCHER "$PATCHER_NAME"
    printf "$format" \
	"WIT"		"$WIT" \
	"WSZST"		"$WSZST" \
	"WIT-VER"	"$WIT_VER" \
	"WSZST-VER"	"$WSZST_VER"
    printf "$sep"
    printf "$format" \
	"SYSTEM"	"$SYSTEM" \
	"MACHINE"	"$MACHINE" \
	"HOST"		"$HOST" \
	"BASEDIR"	"$BASEDIR" \
	"BINDIR"	"$BINDIR" \
	"PATH"		"$PATH"
    printf "$sep"
    printf "$format" \
	"VERBOSE"	"$VERBOSE" \
	"TESTMODE"	"$TESTMODE" \
	"AUTORUN"	"$OPT_AUTORUN" \
	"DISTRIB"	"$OPT_DISTRIB (pack=$OPT_XZ,xz=$OPT_XZ,7z=$OPT_7Z)" \
	"IMAGE-MODE"	"$OPT_IMG_MODE" \
	"IMAGE-RIIV"	"$OPT_IMG_RIIV" \
	"RIIV-MODE"	"$OPT_RIIV_MODE" \
	"ENAB-RIIV"	"$ENABLE_RIIV" \
	"SRC-DIR"	"$SRC_DIR" \
	"DEST-DIR"	"$DEST_DIR" \
	"SPLIT"		"$OPT_SPLIT" \
	"SAVEGAME"	"$OPT_SAVEGAME"
    ((OPT_NEW)) && echo '##' && printf "$format" OLD/NEW $OPT_NEW
  else
    [[ $PATCHER_NAME ]] && printf "$format" PATCHER "$PATCHER_NAME"
    printf "$format" \
	"HOST"		"$HOST" \
	"WIT"		"$WIT" \
	"WSZST"		"$WSZST" \
	"WIT-VER"	"$WIT_VER" \
	"WSZST-VER"	"$WSZST_VER"
  fi
  printf '%s\n' "$sep"
fi

#------------------------------------------------------------------------------
# setup error management

function ON_ERROR()
{
    printf '\n\e[41;37;1m*** SCRIPT FAILED! ***\e[0m\n File: %s\n Line: %u\n' "$1" "$2"
    printf '\e[41;37;1m*** ABORT! ***\e[0m\n'
    exit 99
}

function SETUP_ERROR_HANDLING()
{
    trap 'ON_ERROR "$BASH_SOURCE" "$LINENO"' ERR
}

function CANCEL_ERROR_HANDLING()
{
    trap - ERR
}

export -f ON_ERROR SETUP_ERROR_HANDLING CANCEL_ERROR_HANDLING
SETUP_ERROR_HANDLING

#------------------------------------------------------------------------------
# setup cache directories

export CACHEDIR="$PWD/cache/$REV_UNIQID"

export MKF_PERM_CACHE="$PWD/../mkf-perm-cache"
[[ -d $MKF_PERM_CACHE ]] || MKF_PERM_CACHE="$PWD/mkf-perm-cache"
#echo "MKF_PERM_CACHE=$MKF_PERM_CACHE"

#------------------------------------------------------------------------------
# logging

function printlog_helper()
{
    ((quiet)) && return 0
    echo
    local col="$1"
    shift
    local msg xmsg="$(printf "$@")"
    while read msg
    do
	if [[ $msg = "-" ]]
	then
	    echo
	else
	    local len len1 len2
	    let len=79-$( unset LC_ALL; LC_CTYPE=en_US.UTF-8; echo ${#msg} )
	    ((len<0)) && len=0
	    let len1=len/2
	    let len2=len-len1
	    printf "${col}%*s%s%*s${COL0}\n" $len1 "" "$msg" $len2 ""
	fi
    done <<< "$xmsg"
}

function print_log()
{
    printlog_helper "${COL_LOG}" "$@"
}

function print_sum()
{
    printlog_helper "${COL_SUM}" "$@"
}

function print_activity()
{
    printlog_helper "${COL_ACTIVITY}" "»»» $@ «««"
}

function error_exit() # errcode lines...
{
    local err=$1
    shift

    printlog_helper "${COL_ERROR}" "ERROR $err => ABORT"
    for line in "$@" ---
    do
	printf "\e[31;1m%s\e[0m\n" "$line" >&2
    done
    exit $err
}

#------------------------------------------------------------------------------
# logging, [[obsolete]] 2018-11 => proof usage of other scripts

function printlog()
{
    printlog_helper "${COL_LOG}" "$@"
}

function printsum()
{
    printlog_helper "${COL_SUM}" "$@"
}

#
###############################################################################
###############			configuration			###############
###############################################################################
# setup

ISOMODE_LIST="iso ciso wdf wbfs gcx wia"

CONFIG="./config.def"
CCONFIG="$MKF_PERM_CACHE/config.def"
[[ ! -f $CONFIG && -f "$CCONFIG" ]] && cp "$CCONFIG" "$CONFIG"

CFG_LANGUAGE="$LC_MESSAGE"
[[ $CFG_LANGUAGE = "" ]] && CFG_LANGUAGE="${LC_CTYPE%%_*}"
[[ $CFG_LANGUAGE = "" ]] && CFG_LANGUAGE=en

CFG_MSGLANG1=
CFG_MSGLANG2=

#------------------------------------------------------------------------
# LINE: code(locase) : new_code : mode : language : info
# MODE: A=alias, F=full translated
#------------------------------------------------------------------------
MSGLANG_TAB='
g:G:-F:Deutsch:
u:U:-F:English (America):
e:E:-F:English (Europe):
m:M:-F:Español (América):
s:S:-F:Español (Europa):
q:Q:--:Français (Amérique): => aucun message de chat
f:F:--:Français (Europe):   => aucun message de chat
i:I:--:Italiano:            => messaggi di chat
j:J:--:日本人 (Japanese):   => no chat messages
k:K:--:한국어 (Korean):     => 채팅 메시지, Wiimmfi 문자 없음

#--- aliases
-:-:A-
d:G:AF
de:G:AF
en:E:AF
'

export C_HIGHLIGHT=$'\e'"[33;1m"
export C_HEADING=$'\e'"[33;1m"
export C_TABLE=$'\e'"[37;1m"
export C_ASK=$'\e'"[36;1m"
export C_OFF=$'\e'"[0m"

for d in "$BASEDIR" "$BASEDIR/.." .
do
    LANG_SRC="$d/locale"
    if [[ -d $LANG_SRC ]]
    then
	LANG_LIST="$( cd "$LANG_SRC"; echo *.sh | sed 's/.sh//g' )"
	[[ $LANG_LIST == "*" ]] || break
    fi
done

if [[ -s $CONFIG ]]
then
    eval $( sed '/^#/ d; s/^/CFG_/' "$CONFIG")
fi

if [[ "${CFG_MSGLANG1}{CFG_MSGLANG2}" = "" ]]
then
    case "CFG_MSGLANG" in
	de)	CFG_MSGLANG1=-; CFG_MSGLANG2=G ;;
	de+)	CFG_MSGLANG1=G; CFG_MSGLANG2=G ;;
	en)	CFG_MSGLANG1=-; CFG_MSGLANG2=E ;;
	en+)	CFG_MSGLANG1=E; CFG_MSGLANG2=E ;;
	es)	CFG_MSGLANG1=-; CFG_MSGLANG2=S ;;
	es+)	CFG_MSGLANG1=S; CFG_MSGLANG2=S ;;
	*)	CFG_MSGLANG1=-; CFG_MSGLANG2=  ;;
    esac
fi

[[ $OPT_IMG_SET ]] && CFG_ISOMODE=$OPT_IMG_SET

# [[obsolete]] : CFG_CTREORDER
[[ $CFG_REORDER = "" ]] && CFG_REORDER="$CFG_CTREORDER"
CFG_CTREORDER="$CFG_REORDER"

#
#------------------------------------------------------------------------------

function save_config()
{
    cat <<- --EOT-- >"$CONFIG"
	#-------------------------
	#SYSTEM=$SYSTEM
	#MACHINE=$MACHINE
	#HOST=$HOST
	#-------------------------
	LANGUAGE="$CFG_LANGUAGE"
	MSGLANG1="$CFG_MSGLANG1"
	MSGLANG2="$CFG_MSGLANG2"
	TRACKLANG="$CFG_TRACKLANG"
	CTRENAME="$CFG_CTRENAME"
	REORDER="$CFG_REORDER"
	ISOMODE="$CFG_ISOMODE"
	SPLITISO="$CFG_SPLITISO"
	PRIV_SAVEGAME="$CFG_PRIV_SAVEGAME"
	#-------------------------
	--EOT--

    [[ -d $MKF_PERM_CACHE ]] && cp "$CONFIG" "$CCONFIG"
}

function print_config()
{
    local format=" %-32.32s : %s\n"
    printf "\n----------------------------------------\n"
    printf "$format" "$T_LANGUAGE" "$CFG_LANGUAGE"
    printf "$format" "$T_MSGLANG1" "$CFG_MSGLANG1"
    printf "$format" "$T_MSGLANG2" "$CFG_MSGLANG2"
    printf "$format" "$T_TRACKLANG" "$CFG_TRACKLANG"
    printf "$format" "$T_CTRENAME" "$CFG_CTRENAME"
    printf "$format" "$T_REORDER" "$CFG_REORDER"
    printf "$format" "$T_ISOMODE" "$CFG_ISOMODE"
    printf "$format" "$T_SPLITISO" "$CFG_SPLITISO"
    printf "$format" "$T_PRIV_SAVEGAME" "${CFG_PRIV_SAVEGAME#*=}"
    printf -- "----------------------------------------\n\n"
}

function validate_config()
{
    [[ $CFG_LANGUAGE = "" ]]		&& return 1
    [[ $CFG_MSGLANG1 = "" ]]		&& return 1
    [[ $CFG_MSGLANG2 = "" ]]		&& return 1
    [[ $CFG_ISOMODE = "" ]]		&& return 1
    [[ $CFG_SPLITISO = "" ]]		&& return 1
    [[ $CFG_PRIV_SAVEGAME = "" ]]	&& return 1
    return 0
}

function config_options()
{
    OPT_RIIV_MODE="$ENABLE_RIIV"
    [[ $CFG_ISOMODE ]]			&& setup_image_mode "${CFG_ISOMODE}"
    [[ ${CFG_SPLITISO:0:1} = 1 ]]	&& OPT_SPLIT=1
    [[ ${CFG_PRIV_SAVEGAME:0:1} = 1 ]]	&& OPT_SAVEGAME=1

    if ((VERBOSE>1))
    then
	echo
	printf "\033[32m%10s\033[0m = \033[36m%.70s\033[0m\n" \
		"IMAGE-MODE"	"$OPT_IMG_MODE" \
		"IMAGE-RIIV"	"$OPT_IMG_RIIV" \
		"RIIV-MODE"	"$OPT_RIIV_MODE" \
		"SPLIT"		"$OPT_SPLIT" \
		"SAVEGAME"	"$OPT_SAVEGAME" \
		"LANG_SRC"	"$LANG_SRC" \
		"LANG_LIST"	"$LANG_LIST"
	echo
    fi | sed 's/^/##  /'
}

#
#------------------------------------------------------------------------------
# load_locale()

function load_locale()
{
    local src="$LANG_SRC/$1.sh"
    [[ -s $src ]] && . "$src"
}

load_locale de			# first get german texts
load_locale en			# ... then english texts (both fall back)
load_locale "$CFG_LANGUAGE"	# now scan predefined language

#
#------------------------------------------------------------------------------
# read_char()

function read_char()
{
    local prompt="$1"
    local default="$2"
    shift 2
    while true
    do
	local in
	read -n1 -p "$prompt [$default]: " -s in || exit 1
	if [[ $in = "" ]]
	then
	    echo "$default"
	    echo "$default" >&2
	    return 0
	fi

	for (( i=1; i <= $#; i++ ))
	do
	    local p="${!i}"
	    if [[ $in == ${p:0:1} ]]
	    then
		echo "${p:1}"
		echo "${p:1}" >&2
		return 0
	    fi
	done
	echo >&2
    done
}

#
#------------------------------------------------------------------------------
# read_mode()

function read_mode()
{
    local prompt="$1"
    local default="$2"
    local use_default="$USE_DEFAULT"
    shift 2
    while true
    do
	local in
	if [[ $use_default = 1 ]]
	then
	    use_default=0
	    echo "${C_ASK}$prompt:${C_HIGHLIGHT} $default ${C_OFF}" >&2
	else
	    read -ep "${C_ASK}$prompt [$default]:${C_OFF} " in || exit 1
	fi
	[[ $in = "" ]] && in="$default"
	in="${in,,}"
	local len=${#in}
	local found=
	local count=0
	for (( i=1; i <= $#; i++ ))
	do
	    local p="${!i}"
	    if [[ $in = $p ]]
	    then
		let count=1
		found="$p"
		break
	    elif ((len)) && [[ $in == ${p:0:(len)} ]]
	    then
		let count++
		found="$p"
	    fi
	done
	if (( count == 1 ))
	then
	    echo "$found"
	    [[ -f $PROTOCOL ]] && echo "$prompt:${C_HIGHLIGHT} $found ${C_OFF}" >>"$PROTOCOL"
	    return 0
	fi
    done
}

#
#------------------------------------------------------------------------------
# read_mode()

function read_mode_list()
{
    local prompt="$1"
    local default="$2"
    local use_default="$USE_DEFAULT"
    shift 2
    while true
    do
	local in
	if [[ $use_default = 1 ]]
	then
	    use_default=0
	    echo "${C_ASK}$prompt:${C_HIGHLIGHT} $default ${C_OFF}" >&2
	else
	    read -ep "${C_ASK}$prompt [$default]:${C_OFF} " in || exit 1
	fi
	[[ $in = "" ]] && in="$default"
	in="${in,,}"
	local len=${#in}
	local valid=
	local err=0
	for tag in ${in//,/ }
	do
	    local found=0
	    for (( i=1; i <= $#; i++ ))
	    do
		local p="${!i}"
		if [[ $tag == $p ]]
		then
		    found=1
		    valid="$valid,$tag"
		    break;
		fi
	    done
	    #echo "$found=$tag" >&2
	    ((found)) || let err++
	done
	if [[ $valid != "" ]]
	then
	    if ((!err))
	    then
		echo "${valid#,}"
		[[ -f $PROTOCOL ]] && echo "$prompt:${C_HIGHLIGHT} ${valid#,} ${C_OFF}" >>"$PROTOCOL"
		return 0
	    fi
	    default="${valid#,}"
	fi
    done
}

#
#------------------------------------------------------------------------------
# read_noyes()

function read_noyes() # prompt default allow_quit
{
    local prompt="$1"
    local default="$2"
    local allow_quit="$3"
    [[ ${2:0:1} = 0 ]] && default="$T_NO"
    [[ ${2:0:1} = 1 ]] && default="$T_YES"
    local single="0 1"
    if [[ ${T_NO:0:1} != ${T_YES:0:1} ]]
    then
	single="$single ${T_NO:0:1} ${T_YES:0:1}"
    fi
    [[ $allow_quit = 1 ]] && single="$single quit"

    local answer="$(read_mode "$prompt ($T_NO,$T_YES)" "$default" "$T_NO" "$T_YES" $single)"
    if [[ "$answer" = $T_NO || "$answer" = 0 ]]
    then
	answer="0=$T_NO"
    elif [[ "$answer" = $T_YES || "$answer" = 1 || "$answer" = ${T_YES:0:1} ]]
    then
	answer="1=$T_YES"
    elif [[ "$answer" != quit ]]
    then
	answer="0=$T_NO"
    fi

    echo "$answer"
    return 0
}

#
#------------------------------------------------------------------------------
# ask_lang()

function ask_lang() # step config ask
{
    local step="$1"
	  cfg="$2"
    local ask="$3"

    cfg=$( awk -F: "\$1 == \"${cfg,,}\" { print \$2; exit }" <<< "$MSGLANG_TAB")

    local x_list=""
    local x_info=""
    while IFS=: read in code mode lang info
    do
	#echo "| $in | $code | $mode | $lang | $info |"
	[[ $in =~ ^# || $in = "" || $code = "" ]] && continue
	[[ $step = 2 && ! $mode =~ F ]] && continue

	x_list="${x_list} $in"
	if [[ ! $mode =~ A ]]
	then
	    printf "   %s : %s %s\n" "$code" "$lang" "$info"
	    x_info="${x_info},$code"
	fi
    done <<< "$MSGLANG_TAB"
    x_list="${x_list:1}"
    x_info="${x_info:1}"
    [[ $step = 1 ]] && x_info="${x_info},-"

    echo
    cfg="$(read_mode "* $ask ($x_info)" "$cfg" $x_list )"
    cfg=$( awk -F: "\$1 == \"${cfg,,}\" { print \$2; exit }" <<< "$MSGLANG_TAB")
    #echo "CFG=$cfg"
}

#
#------------------------------------------------------------------------------
# ask_config_sub()

function ask_config_sub()
{
    #----- dialog language

    if [[ $OPT_AUTORUN != 1 && ${ASK_LANGUAGE:=1} = 1 ]]
    then
	printf "$T_LANGUAGE_INFO\n"
	CFG_LANGUAGE="$(read_mode "* $T_LANGUAGE (${LANG_LIST// /,})" \
		"$CFG_LANGUAGE" $LANG_LIST wim wiimm )"

	if [[ $CFG_LANGUAGE = wim || $CFG_LANGUAGE = wiimm ]]
	then
	    CFG_LANGUAGE=de
	    CFG_MSGLANG1=G
	    CFG_MSGLANG2=G
	    CFG_TRACKLANG=wiimm
	    CFG_LANGUAGE="$(read_mode "* $T_LANGUAGE (${LANG_LIST// /,})" \
		"$CFG_LANGUAGE" $LANG_LIST)"
	fi
    fi
    load_locale "$CFG_LANGUAGE"


    #----- game language

    if [[ $OPT_AUTORUN != 1 && ${ASK_MSGLANG:=1} = 1 ]]
    then
	printf "$T_MSGLANG1_INFO\n"
	ask_lang 1 "$CFG_MSGLANG1" "$T_MSGLANG1"
	CFG_MSGLANG1="${cfg^^}"

	x=$( awk -F: "\$1 == \"${cfg,,}\" { print; exit }" <<< "$MSGLANG_TAB")
	IFS=: read in code mode lang info <<< "$x"
	if [[ $mode =~ F ]]
	then
	    CFG_MSGLANG2=-
	else
	    printf "$T_MSGLANG2_INFO\n"
	    ask_lang 2 "${CFG_MSGLANG2%-}" "$T_MSGLANG2"
	    CFG_MSGLANG2="${cfg^^}"
	fi
    fi
    export CFG_MSGLANG1
    export CFG_MSGLANG2


    #----- track language

    if [[ $OPT_AUTORUN != 1 && ${ASK_TRACKLANG:=1} = 1 && -f ./track-language.list ]]
    then
	printf "$T_TRACKLANG_INFO\n"
	[[ $CFG_TRACKLANG = "" ]] && CFG_TRACKLANG="x,$CFG_LANGUAGE"

	local x_list="x"
	while IFS=: read id id2 code locale info
	do
	    #echo "| $id | $id2 | $code | $locale | $info |"
	    [[ $id =~ ^# || $id = "" || $info = "" ]] && continue

	    x_list="${x_list} $id"
	    printf "   %s : %s\n" "$id" "$info"
	done <<< "$(cat ./track-language.list)"
	echo

	CFG_TRACKLANG="$(read_mode_list "* $T_TRACKLANG" "$CFG_TRACKLANG" $x_list wim wiimm )"
	if [[ $CFG_TRACKLANG = wim || $CFG_TRACKLANG = wiimm ]]
	then
	    CFG_TRACKLANG=en
	    CFG_CTRENAME=0
	    CFG_REORDER=0
	    #---
	    [[ $REV_REORDER < 1 ]] && CFG_REORDER=0
	    USE_DEFAULT=1
	    CFG_TRACKLANG="$(read_mode_list "* $T_TRACKLANG" "$CFG_TRACKLANG" $x_list )"
	fi

	if [[ $REV_RENAME < 2 ]]
	then
	    CFG_CTRENAME=0
	else
	    CFG_CTRENAME="$(read_noyes "* $T_CTRENAME" "$CFG_CTRENAME")"
	fi

	if [[ $REV_REORDER < 1 ]]
	then
	    CFG_REORDER=0
	else
	    CFG_REORDER="$(read_noyes "* $T_REORDER" "$CFG_REORDER")"
	fi
	USE_DEFAULT=0
    fi
    export CFG_TRACKLANG
    export CFG_CTRENAME
    export CFG_REORDER
    # [[obsolete]] : CFG_CTREORDER
    export CFG_CTREORDER="$CFG_REORDER"


    #----- iso mode (or Riivolution)

    [[ -d $BASEDIR/riivolution ]] || ENABLE_RIIV=0

    local info="$T_ISOMODE_INFO"
    local list="$ISOMODE_LIST"
    if [[ $ENABLE_RIIV = 1 ]]
    then
	info="${info}${T_ISOMODE_RIIV1}"
	list="$list riiv"
    elif [[ $ENABLE_RIIV = 2 ]]
    then
	info="${info}${T_ISOMODE_RIIV2}"
	list="$list riiv riiv-"
    elif (( ENABLE_RIIV > 2 ))
    then
	info="${info}${T_ISOMODE_RIIV3}"
	list="$list riiv"
    fi

    if [[ $OPT_AUTORUN != 1 && ${ASK_ISOMODE:=1} = 1 ]]
    then
	printf "$info\n"
	CFG_ISOMODE="$(read_mode "* $T_ISOMODE (${list// /,})" "$CFG_ISOMODE" $list none)"
    fi

    CREATE_RIIV=0
    [[ $CFG_ISOMODE = riiv ]] && CREATE_RIIV=1
    [[ $CFG_ISOMODE = riiv- ]] && CREATE_RIIV=2


    #----- split image

    if [[ $OPT_AUTORUN != 1 && ${ASK_SPLITISO:=1} = 1 && $CREATE_RIIV = 0 ]]
    then
	printf "$T_SPLITISO_INFO\n"
	CFG_SPLITISO="$(read_noyes "* $T_SPLITISO" "$CFG_SPLITISO")"
    fi


    #----- alternate save game

    if [[ $OPT_AUTORUN != 1 && ${ASK_SAVEGAME:=1} = 1  && $CREATE_RIIV = 0 ]]
    then
	printf "$T_SAVEGAME_INFO\n"
	CFG_PRIV_SAVEGAME="$(read_noyes "* $T_PRIV_SAVEGAME" "$CFG_PRIV_SAVEGAME")"
    fi


    #----- save config

    save_config
}

#
#------------------------------------------------------------------------------
# ask_config()

function ask_config()
{
    while true
    do
	export PROTOCOL="$LOGDIR/settings.txt"
	rm -f "$PROTOCOL"
	touch "$PROTOCOL"

	ask_config_sub "$@"
	[[ -s $PROTOCOL ]] || break

	printf "\033[H\033[J\n"
	printf "$T_SUMMARY_INFO\n"
	cat "$PROTOCOL"
	echo
	[[ $OPT_AUTORUN = 1 ]] && break

	cont="$(read_noyes "* $T_SUMMARY" "$T_YES" 1 )"
	[[ $cont = quit ]] && exit 1
	[[ $cont =~ ^1 ]] && break
    done

    config_options
    return 0
}

#
###############################################################################
###############			create image			###############
###############################################################################

DONE_COUNT=0
DISTRIB_DONE=0

#------------------------------------------------------------------------------

function create_distrib()
{
    if (( OPT_DISTRIB && !DISTRIB_DONE ))
    then
	DISTRIB_DONE=1

	local job=../create-distrib.sh
	[[ -x $job ]] || job=../../create-distrib.sh
	if [[ -x $job ]]
	then
	    local opt=()
	    ((OPT_DB_DISTRIB)) && opt+=(--db)
	    ((OPT_PACK)) && opt+=(--pack)
	    ((OPT_RSYNC)) && opt+=(--rsync=$OPT_RSYNC)
	    if ((OPT_7Z>1))
	    then
		opt+=( --7z --xz --xz )
	    else
		((OPT_XZ)) && opt+=( --xz --xz )
		#((OPT_XZ>1)) && opt+=(--xz)
		((OPT_7Z)) && opt+=(--7z)
	    fi

	    if [[ -a /dev/fd/3 ]]
	    then
		"$job" "${opt[@]}" 2>&3
	    else
		"$job" "${opt[@]}"
	    fi
	fi
	if (( OPT_DISTRIB > 1 ))
	then
	    let DONE_COUNT++
	    OPT_IMG_MODE=none
	fi
    fi
}

#
#------------------------------------------------------------------------------
# function add2filename()

function add2filename()
{
    local fn="$1"
    local add="$2"
    local dir=

    if [[ $fn =~ / ]]
    then
	dir="${fn%/*}/"
	fn="${fn##*/}"
    fi

    if [[ $fn =~ \. ]]
    then
	echo "${dir}${fn%.*}${add}.${fn##*.}"
    else
	echo "${dir}${fn}${add}"
    fi
}

#
#------------------------------------------------------------------------------
# function extract_patch()

function extract_patch()
{
    local patchfile="$1"
    local workdir="$2"

    print_log "$T_LOG_PATCH" "$patchfile"

    if [[ -d $patchfile ]]
    then
	[[ -d $patchfile/patch-dir ]] && patchfile+=/patch-dir
	mkdir -p "$workdir/"
	local cpmode=
	((VERBOSE>1)) && cpmode=-v
	$TEST cp $cpmode -rl --remove-destination --dereference "$patchfile"/* "$workdir/"
    elif [[ ${patchfile##*.} == zip ]]
    then
	local zipmode=-q
	((VERBOSE>1)) && zipmode=
	$TEST unzip -o $zipmode "$patchfile" \
		|| error_exit $ERR_ERROR "unzip failed: $patchfile"
    else
	local tarmode=xf
	((VERBOSE>1)) && tarmode=xvf
	$TEST tar $tarmode "$patchfile" --no-same-owner \
		|| error_exit $ERR_ERROR "un-tar failed: $patchfile"
    fi
}

#
#------------------------------------------------------------------------------
# function job_mkw()

function job_mkw()
{
    local start_time="$(date +'%F %T/%s')"
    create_distrib

    local SED_TITLE='
	s|/?@?@([A-Z0-9-]+)@?@||g;
	s|/+|-|g;
	s|\s+| |g;
	s|^\s+||;
	s|\s+$||;
    '

    local title="$REV_TITLE"
    [[ $REV_VERSION = "" ]] || title="$REV_TITLE.$REV_VERSION"
    title="$( sed -r "$SED_TITLE" <<<"$title" )"

    # job source_id dest_id "title" patch_file
    job RMCP01 "RMCP${REV_SUBID}" "$title.pal" patch.tar || return 1
    job RMCE01 "RMCE${REV_SUBID}" "$title.usa" patch.tar || return 1
    job RMCJ01 "RMCJ${REV_SUBID}" "$title.jap" patch.tar || return 1
    job RMCK01 "RMCK${REV_SUBID}" "$title.kor" patch.tar || return 1

    local end_time="$(date +'%T/%s')"
    local sec=$(( ${end_time#*/} - ${start_time#*/} ))
    if ((sec>99))
    then
	sec="$( printf "%um+%us" $((sec/60)) $((sec%60)) )"
    else
	sec="${sec}s"
    fi

    print_sum "$T_LOG_DONE_COUNT\n%s .. %s (%s)\n-" \
		"$DONE_COUNT" "$OPT_IMG_MODE" \
		"${start_time%/*}" "${end_time%/*}" "$sec"
    return 0
}

#
#------------------------------------------------------------------------------
# function load_option_file()

function load_option_file()
{
    local wstrt_rel="$X_WSTRT_REL"
    local wstrt_del="$X_WSTRT_DOL"
    export X_WSTRT_REL= # for staticR.rel
    export X_WSTRT_DOL= # for main.dol

    if [[ $1 ]]
    then
	if [[ -s $1 ]]
	then
	    . $1
	elif ! [[ $1 =~ / ]]
	then
	    if [[ -s "../$1" ]]
	    then
		. "../$1"
	    elif [[ -s "../../$1" ]]
	    then
		. "../../$1"
	    fi
	fi
    fi

    # restore options
    export X_WSTRT_REL="$wstrt_rel"
    export X_WSTRT_DOL="$wstrt_del"
}

#
#------------------------------------------------------------------------------
# function job()

function job()
{
    create_distrib
    (( DONE_COUNT )) && [[ $OPT_IMG_MODE = none ]] && return 0
    (( OPT_IMG_RIIV && DONE_COUNT )) && return 0

    if (( OPT_IMG_RIIV > 1 ))
    then
	job_riiv2 "$@"
	return $?
    fi

    local srcid="$1"
    local destid="$2"
    local destname="$3"
    local patchfile="$4"
    export LANG_ID="${srcid:3:1}"

    ((OPT_SAVEGAME)) && destname="${destname}+"


    #----- extract

    print_log "$T_LOG_LIST_SOURCE" "$srcid"
    n=0

    rm -rf "$workdir/sys-save"
    [[ -d "$workdir/sys" ]] && mv "$workdir/sys" "$workdir/sys-save"
    rm -rf "$workdir/sys"

    if [[ -f $SRC_DIR ]]
    then
	SRC_FILE="$SRC_DIR"
	SRC_DIR=
	n=$( wit filetype --no-header --include $srcid "$SRC_FILE" | wc -l )
	((VERBOSE>0)) && printf "source file, id=%s, OK=%d: %s\n" $srcid $n "$SRC_FILE"
	((n)) || return 0
    elif [[ -d $SRC_DIR ]]
    then
	nn=$( wit filetype --no-header "$SRC_DIR" | wc -l )
	n=$( wit filetype --no-header --include $srcid "$SRC_DIR" | wc -l )
	((VERBOSE>0)) && printf "source dir, id=%s, N=%d: %s\n" $srcid $n "$SRC_DIR"
	((!n&&nn)) && return 0
    fi

    [[ -d "$workdir/sys-save" ]] && mv "$workdir/sys-save" "$workdir/sys"

    if ((!n))
    then
	n=$( wit filetype --no-header --include $srcid | wc -l )
	((n)) || return 0
	SRC_DIR=()
    fi

    if ((VERBOSE>=0))
    then
	if [[ $SRC_FILE = "" ]]
	then
	    wit filetype --long "${SRC_DIR[@]}" . --include $srcid
	else
	    wit filetype --long "$SRC_FILE" . --include $srcid
	fi
    fi

    ((VERBOSE<0)) && echo "** Create $destname"
    let DONE_COUNT++

    print_log "$T_LOG_RM_WORKDIR"
    local extractdir="$workdir"
    $TEST rm -rf "$extractdir"

    zero=""
    [[ -s ./zero-files.list ]] && zero+=' --zero-files=@./zero-files.list'
    [[ -s ./ignore-files.list ]] && zero+=' --ignore-files=@./ignore-files.list'

    export OWN_FILES=0
    if [[ -f $SRC_FILE ]]
    then
	print_log "$T_LOG_EXTRACT_IMAGE" "$srcid"
	$TEST wit extract $WIT_VERBOSE -1p "$SRC_FILE" --include $srcid \
		$zero --DEST "$extractdir" --psel data \
		|| error_exit $ERR_ERROR "wit extract failed: $SRC_FILE"
	OWN_FILES=1
    elif [[ -d $SRC_DIR ]]
    then
	print_log "$T_LOG_COPY_IMAGE" "$srcid"
	# protect 'SRC_DIR' against 'tar extract'
	extractdir="$workdir.extract"
	$TEST rm -rf "$extractdir"
	$TEST cp -rl --remove-destination --dereference "$SRC_DIR" "$extractdir"
    elif ((OPT_IMG_RIIV)) && [[ -f ./riiv-extract.list ]]
    then
	print_log "$T_LOG_EXTRACT_RIIV" "$srcid"
	$TEST wit extract $WIT_VERBOSE -1p . --include $srcid \
		--files=@./riiv-extract.list \
		--DEST "$extractdir" --psel data \
		|| error_exit $ERR_ERROR "wit extract failed"
	OWN_FILES=1
    else
	print_log "$T_LOG_EXTRACT_IMAGE" "$srcid"
	$TEST wit extract $WIT_VERBOSE -1p . --include $srcid \
		$zero --DEST "$extractdir" --psel data  \
		|| error_exit $ERR_ERROR "wit extract failed"
	OWN_FILES=1
    fi

    if  ((OPT_IMG_RIIV))
    then
	printf "$T_LOG_RIIV_TIMESTAMP"
	find "$extractdir" -type f -exec touch -t 200001010000 {} \;
    fi


    #----- pre-patch

    if ((!TESTMODE)) && [[ -f "./pre-patch.sh" ]]
    then
	print_log "$T_LOG_PATCH" "pre-patch.sh"
	echo
	ln -f "./pre-patch.sh" "$extractdir/"
	( cd "$extractdir"; bash ./pre-patch.sh $fast ) || exit 1
    fi


    #----- extract patch file/dir

    extract_patch "$patchfile" "$workdir"
    patchfile2=$(add2filename "$patchfile" 2)
    [[ -a $patchfile2 ]] && extract_patch "$patchfile2" "$workdir"

    if ((OPT_PATCH3))
    then
	patchfile3=$(add2filename "$patchfile" 3)
	[[ -a $patchfile3 ]] && extract_patch "$patchfile3" "$workdir"
    fi

    [[ -f "$workdir/post-extract.sh" ]] && . "$workdir/post-extract.sh"

    if [[ $workdir != $extractdir ]]
    then
	cp -rl --remove-destination --dereference "$workdir"/* "$extractdir/"
	rm -rf "$workdir"
	mv "$extractdir" "$workdir"
    fi
    #((TESTMODE)) || [[ $HOST = cygwin ]] || chmod -R a+Xrw "$workdir"

    rm -f "$workdir/setup.txt"
    printf "part-id = %s\npart-name = %s\n" \
			"$destid" "$destname" > "$workdir/setup.txt"

    local opt_id=(--id "$destid")
    if ((OPT_SAVEGAME))
    then
	opt_id=(--id=K --modify boot,tmd,ticket)
	if [[ -s "$workdir/savebanner.tpl" && -d "$workdir/files/Boot/" ]]
	then
	    echo " * use alternate 'savebanner.tpl'"
	    ln -f "$workdir/savebanner.tpl" "$workdir/files/Boot/"
	fi
    fi


    #----- OPTION_FILE

    load_option_file "$OPTION_FILE"


    #----- post-patch.sh

    if ((!TESTMODE)) && [[ -f "$workdir/post-patch.sh" ]]
    then
	echo
	( cd "$workdir"; bash ./post-patch.sh $fast ) || exit 1
	if [[ -d "$workdir/_Wiimm" ]]
	then
	    rm -rf ./_Wiimm
	    cp -rl "$workdir/_Wiimm" ./
	fi

    fi
    #((TESTMODE)) || [[ $HOST = cygwin ]] || chmod -R a+Xrw "$workdir"


    #----- compose a new image

    if [[ $DEST_FILE = "" ]]
    then
	if [[ $OPT_IMG_MODE == wbfs ]]
	then
	    DEST_FILE=./new-image/%Y/%+
	else
	    DEST_FILE=./new-image/%X
	fi
    elif [[ -d $DEST_FILE || ${DEST_FILE##*/} = "" ]]
    then
	DEST_FILE="$DEST_FILE/%X"
    fi

    local opt_base=(--links --overwrite --titles=0 $WIT_VERBOSE "${opt_id[@]}")

    if ((OPT_IMG_RIIV))
    then
	job_riiv1 "$@"
    elif  [[ $OPT_IMG_MODE = none ]]
    then
	true
    elif  [[ $OPT_IMG_MODE = add ]]
    then
	print_log "$T_LOG_ADD_IMAGE" "$destid"
	$TEST wwt --all "${opt_base[@]}" add "$workdir" || exit 1
    else
	print_log "$T_LOG_CREATE_IMAGE" "$destid" "./new-image/*.$OPT_IMG_MODE"
	$TEST wit copy "${opt_base[@]}" "$workdir" --DEST "$DEST_FILE" \
		--$OPT_IMG_MODE "${WIT_SPLIT[@]}" || exit 1
    fi
}

#
#------------------------------------------------------------------------------
# function job_riiv1()

function job_riiv1()
{
    # DONE_COUNT already incremented
    setup_riiv "$@"
    ((TESTMODE)) && return 0
    printf "$T_LOG_RIIV_TRANSFER"
    find "$workdir/files" -depth -name '.?*' -exec rm -rf {} \;
    ( cd "$workdir/files" && find . -type f -mtime -3650 ) | \
	while read f
	do
	    dir="${f%/*}"
	    mkdir -p "$FDIR/$dir"
	    ln -f "$workdir/files/$f" "$FDIR/$f"
	done
    term_riiv_mkw "$@"
}

#
#------------------------------------------------------------------------------
# function job_riiv2()

function job_riiv2()
{
    let DONE_COUNT++
    setup_riiv "$@"

    case "$OPT_RIIV_MODE" in
	2) term_riiv_mkw "$@" ;;
	4) term_riiv_xml "$@" ;;
	5) term_riiv_script "$@" ;;
	*) term_riiv_other "$@" ;;
    esac
}

#
#------------------------------------------------------------------------------
# function setup_riiv()

function setup_riiv()
{
    local patchfile="$4"

    RIIV="./riiv-sd-card"
    if [[ $1 == "" ]]
    then
	mkdir -p "$RIIV"
	RIIV="$RIIV/$1"
    fi
    print_log "$T_LOG_RIIV_START" "$RIIV"
    printf "$T_LOG_RIIV_SETUP"

    rm -rf "$RIIV"
    cp -rl "$RIIV_SRC" "$RIIV"
    #rm -rf "$RIIV"/.svn*
    mkdir -p "$RIIV/riivolution/config"

    TITLE="$REV_TITLE"
    [[ $TITLE = "" ]] && TITLE="$REV_NAME"
    [[ $TITLE = "" ]] && TITLE="MKW-$(date +%F)"
    NAME="$REV_NAME"
    [[ $NAME = "" ]] && NAME="$TITLE"
    DIR="${REV_ID//[!A-Za-z0-9_-]}"
    [[ $DIR = "" ]] && DIR="MKW-$(date +%F)"
    ID="${REV_ID//-/}"
    ID3="${ID//-/}"
    SAVEDIR="riivolution/save/${DIR%%-*}"

    FDIR="$RIIV/$DIR"
    mkdir -p "$FDIR" "$RIIV/$SAVEDIR"

    if [[ -d $patchfile ]]
    then
	local cpmode=
	((VERBOSE>1)) && cpmode=-v
	local cpdest="$RIIV/${workdir##*/}/"
	mkdir -p "$cpdest"
	$TEST cp $cpmode -r --preserve=timestamps "$patchfile"/* "$cpdest"
    elif [[ ${patchfile##*.} == zip ]]
    then
	local zipmode=-q
	((VERBOSE>1)) && zipmode=
	( cd "$RIIV" && $TEST unzip -o $zipmode "../$patchfile" ) || exit 1
    else
	local tarmode=xf
	((VERBOSE>1)) && tarmode=xvf
	$TEST tar $tarmode "$patchfile" --no-same-owner -C "$RIIV" || exit 1
    fi

    $TEST mv "$RIIV/$workdir"/files/* "$FDIR"
    rm -rf "$RIIV/$workdir" "$FDIR/_Wiimm"
}

#
#------------------------------------------------------------------------------
# function term_riiv_mkw()

function term_riiv_mkw()
{
    find "$RIIV" -depth -type d -name '.?*' -exec rm -rf {} \;
    rm -rf "$FDIR/_Wiimm"
    printf "$T_LOG_RIIV_CREATE" "$DIR.xml"
    {
	patch_xml "$RIIV/template/mario-kart-wii-1.xml"
	( cd "$FDIR" && riiv_files "$DIR" Scene/UI Boot/Strap rel \
		Race/Course Race/Common thp/course )
	[[ -d "$FDIR/sys" ]] && patch_xml "$RIIV/template/mario-kart-wii-2.xml"
	patch_xml "$RIIV/template/mario-kart-wii-3.xml"
	( cd "$FDIR" && riiv_files "$DIR" Race/Kart )
	patch_xml "$RIIV/template/mario-kart-wii-4.xml"

    } >"$RIIV/riivolution/$DIR.xml"
    rm -rf "$RIIV/template"
    [[ $HOST = cygwin ]] || chmod -R a+Xrw "$RIIV"
}

#
#------------------------------------------------------------------------------
# function term_riiv_other()

function term_riiv_other()
{
    rm -rf "$FDIR/_Wiimm"
    printf "$T_LOG_RIIV_CREATE" "$DIR.xml"
    {
	patch_xml "$RIIV/template/other-1.xml"
	( cd "$FDIR" && riiv_files "$DIR" . )
	patch_xml "$RIIV/template/other-2.xml"

    } >"$RIIV/riivolution/$DIR.xml"
    rm -rf "$RIIV/template"
}

#
#------------------------------------------------------------------------------
# function term_riiv_xml()

function term_riiv_xml()
{
    rm -rf "$FDIR/_Wiimm" "$RIIV/template"
    cp -p riivolution.xml "$RIIV/riivolution/$DIR.xml"
}

#
#------------------------------------------------------------------------------
# function term_riiv_script()

function term_riiv_script()
{
    rm -rf "$FDIR/_Wiimm"
    [[ -f riivolution.xml ]] && cp -p riivolution.xml "$RIIV/riivolution/$DIR.xml"
    DIR="$DIR" FDIR="$FDIR" RIIV="$RIIV" bash riivolution.sh "$@"
    rm -rf "$RIIV/template"
}

#
#------------------------------------------------------------------------------
# function copy_xml()

function patch_xml()
{
   local src="$1"
   if ((REV_MODE>=200))
   then
	local le="${src%.xml}-le.xml"
	[[ -s "$le" ]] && src="$le"
   fi

   sed "s|@@ID3@@|${destid:0:3}|g;
	s|@@ID4@@|${destid:0:4}|g;
	s|@@ID@@|$ID|g;
	s|@@NAME@@|$NAME|g;
	s|@@TITLE@@|$TITLE|g;
	s|@@DIR@@|$DIR|g; s|@@SAVEDIR@@|$SAVEDIR|g
	" "$src"
}

#
#------------------------------------------------------------------------------
# function riiv_files()

function riiv_files()
{
    local format="$(printf \
	'  <file disc="/%%s" external="/%s/%%s" create="true" />\\n' "$1" )"
    local list=".list.tmp"
    rm -f "$list"
    shift

    local dlist=()
    for d in "$@"
    do
	[[ -d $d ]] && dlist=( "${dlist[@]}" "$d" )
    done
    (( ${#dlist[@]} )) || return 0

    find "${dlist[@]}" -type f -printf '%s\n' | sort -n | uniq |
     while read size
     do
	find "${dlist[@]}" -type f -size ${size}c | sed 's+^\./++' | sort >"$list"
	while [[ -s $list ]]
	do
	    cat /dev/null >"$list.new"
	    base=
	    while read file
	    do
		if [[ $base = "" ]]
		then
		    base="$file"
		    printf "$format" "$file" "$base"
		elif diff -q "$file" "$base" >/dev/null
		then
		    rm -f "$file"
		    printf "$format" "$file" "$base"
		else
		    echo "$file" >>"$list.new"
		fi
	    done <"$list"
	    mv "$list.new" "$list"
	done
    done | sort

    rm -f "$list"
}

#
###############################################################################
###############			    E N D			###############
###############################################################################
